/*  corrEigen.c  - Oct '98 */
/*  VERSION 0.0:
    Input set of eigen basis images,
     compute expansion of given image,
     output float image of residuals.

     Usage:
      corrEigen searchImages.pgm
     prompts for input of eigen basis set
     and output pathname

     Writes sum of squared error image to pathname+imageName_err.pfm
     and coeffs to stdout.
  
******************/

#include  "macros.h"
#include  "utils.h"
#include  "imageFile-io.h"

int debug = 0;  /* 0 for no debug info , > 0 for some debug output */

unsigned char grey = 127;
int eyeRadX = 10;  
int eyeUpY = 15;           /* Need this to find eye center in basis */
int eyeDownY = 10;

void corrEigenBasis(float *residual, float *logLikeImage, float minLogLike,
		   float *meanImage,
                   float *bases[], int nxBasis, int nyBasis, int nBasis,
		   float *sv,
                   int xCent, int yCent,
                   unsigned char *image, int nxImage, int nyImage,
                   float *coeff)
{
  float *diffImage, logLikeEst;
  float mean, var, meanDebug, varDebug, tmp, scl, maxMeanImage, range[2];
  int ix, iy, ixc, iyc, ixy, ixyc, j, k, countDebug, debugMessage;

  grabFloatMemory(&diffImage, nxBasis * nyBasis, "diffImage");
  countDebug = debug;

  maxMeanImage = MACHINE_EPS;
  range_float_image(meanImage, nxBasis, nyBasis, range);
  /** 
  fprintf(stderr, "DEBUG: range mean image: %e %e\n", range[0], range[1]);
  **/
  maxMeanImage = max(maxMeanImage, abs(range[0])); 
  maxMeanImage = max(maxMeanImage, abs(range[1])); 

  for(iyc=0; iyc<nyImage-nyBasis+1; iyc++) {
    for(ixc=0; ixc<nxImage-nxBasis+1; ixc++) {

      debugMessage = 0;

      /* Get mean and power in patch (with top left corner at ixc,iyc),
         and compute inner product with mean-DC image.  Here it is
         important that the mean image has a zero DC component. */
      mean = var = scl = 0.0;
      for(iy=iyc; iy<iyc+nyBasis; iy++)
	for(ix=ixc; ix<ixc+nxBasis; ix++) {
	  tmp = (float) (image[index(ix,iy,nxImage)] - (int) grey);
	  mean += tmp;
	  var += tmp * tmp;
	  scl += image[index(ix,iy,nxImage)] * 
	    meanImage[index(ix-ixc, iy-iyc, nxBasis)];
        }
      mean /= nxBasis * nyBasis;
      mean += grey;
      var -= (mean - grey) * (mean - grey) * nxBasis * nyBasis;;

      /* What to do when scl is <= 0?  
	   scl < 0 is a negative mean image
	   scl = 0... no correlation with the mean image  */

      scl = max(scl, 1.0/maxMeanImage); 
      /* This minimum scl size corresponds to only a +-1 grey-level variation
	in the component of the mean image.  Any smaller is not resolvable.
        NOTE: Image negatives are not considered as scl must be >0 here. */

      /* Compute difference from mean image (and no DC) */
      meanDebug = varDebug = 0.0;
      for(iy=iyc; iy<iyc+nyBasis; iy++)
	for(ix=ixc; ix<ixc+nxBasis; ix++) {
	  ixyc = index(ix-ixc, iy-iyc, nxBasis);
	  tmp = image[index(ix,iy,nxImage)] - mean - scl*meanImage[ixyc];
	  diffImage[ixyc] = tmp;
	  meanDebug += tmp;
	  varDebug += tmp * tmp;
	}
      meanDebug /= nxBasis * nyBasis;
      varDebug -= meanDebug * meanDebug * nxBasis * nyBasis;
      var = varDebug;
 
      /* Compute residual */
      ixy = index(ixc+xCent, iyc+yCent, nxImage);
      residual[ixy] = var;
      tmp = 2.0 * scl * scl;
      logLikeImage[ixy] = 0.0;
      logLikeEst = -var / (tmp * sv[0] * sv[0]);

      /* Project out eigen directions, keeping track of residual
	 variance and logLikelihood.  The log likelihood estimate
         is monotonic decreasing in k, the number of basis functions,
	 and we use this to break out of the loop when the log likelihood
	 gets too small. */
      k = -1;
      if (var > 0.0 && logLikeEst > minLogLike) {
	/* Compute expansion coeffs in patch (top left corner at ixc,iyc) */
        if (countDebug > 0 && k == 0) {
	  fprintf(stderr, "starting: sv[0] %e var %e scl %e : logLikeEst %e\n",
		  sv[0], var, scl, logLikeEst);
	  debugMessage = 1;
	}
	for(k=0; k<nBasis; k++) {
	  coeff[k] = 0.0;
	  for(j=0; j<nxBasis*nyBasis; j++)
	    coeff[k] += bases[k][j]*diffImage[j]; 
          residual[ixy] -= coeff[k] * coeff[k];
	  logLikeImage[ixy] -= coeff[k] * coeff[k]/(tmp * sv[k] * sv[k]);
	  logLikeEst = logLikeImage[ixy] 
	                - residual[ixy]/(tmp * sv[k+1] * sv[k+1]);
	  if (countDebug > 0) {
    
	    fprintf(stderr, "basis %d: coeff %e logLikeEst %e\n", 
		    k, coeff[k], logLikeEst);
	    debugMessage = 1;
          }
	  if (logLikeEst < minLogLike)
	    break;  /* Don't bother with remaining eigen components,
		       it just gets more unlikely */
        }
      }
      logLikeImage[ixy] = logLikeEst;

      /* Scale the residual by the power in the patch, so
	 the residual is in [0,1] for output.  Then 
	 residual = 0 denotes 100% of the variance
	 is accounted for by the eigenmodel, 
	 = 1 denotes 0% */
      residual[ixy] /= var; 

      if (countDebug > 0) countDebug -= debugMessage;
     
    } /* over ixc */
  } /* over iyc */
  free(diffImage);
}

int main(int argc, char **argv)
{
  float *basisImages[MAXLEN], *residualImage, *logLikeImage; 
  float *imageFlt, *meanImage;
  float mean, *coeff, *sv;
  float sumRes, rmsRes, rmsSignal;
  float crop[2], range[2], out_of_range[2], quant_par[2];
  float facePos[6];
  float minLogLike;

  int nx, ny, nImage, ixy;
  int kBasis, nBasis, nxBasis, nyBasis;
  int xCent, yCent;
  int nQuad, nxImage,nyImage;
  int faceCoords, use_mean;
  int iShift, iChannel;
  int size, i,k, c, kImage,start;

  unsigned char *image;
  char fnBasis[MAXLEN][MAXLEN];
  char fnImagePath[MAXLEN][MAXLEN], fnImage[MAXLEN][MAXLEN];
  char fnInput[MAXLEN];
  char outName[MAXLEN];
  char pathNameInput[MAXLEN];
  char pathNameOutput[MAXLEN], outputFileRoot[MAXLEN];
  char rootName[MAXLEN], numString[MAXLEN],scrName[MAXLEN];
  char fileType[3], *pc;
  FILE *fpIn;

  fpIn = NULL;
  nImage = argc-1;
  if (nImage==0) {
    fprintf(stderr, "usage: corrEigen {searchImages.pgm}\n");
    exit(0);
  }
  fprintf(stderr," Number of test images = %d\n", nImage);
  /* Copy filename args to fnImagePath, fnImage */
  for(kImage=0; kImage<nImage; kImage++) {
    if ((pc = strrchr(*++argv, '/')) != NULL) {
      strcpy(fnImagePath[kImage], *argv);
      strcpy(fnImage[kImage], pc+1);
      pc = strrchr(fnImagePath[kImage], '/');
      *(pc+1) = '\0';
    } else {
      strcpy(fnImage[kImage], *argv);
      strcpy(fnImagePath[kImage],"");
    }
  }

  fprintf(stderr," enter minimum log Likelihood to consider (eg. -4): ");
  scanf("%e", &minLogLike);
  fprintf(stderr," enter output file pathname and rootname: ");
  scanf("%s %s",pathNameOutput, outputFileRoot);

  /*******************/
  /* Read basis data */
  /*******************/
  fprintf(stderr," enter basis file pathname : ");
  scanf("%s",pathNameInput);
  /*
  fprintf(stderr, "Input mean image (or \"null\" if no mean image):\n");
  */
  fprintf(stderr, "Input mean image:\n");
  if (scanf("%s", rootName) != EOF) {
    if (checkForNull(rootName) == 0) {
      fprintf(stderr, "Mean image assumed to be zero.\n");
      use_mean = FALSE;
    } else {
      use_mean = TRUE;
      strcpy(fnInput, pathNameInput);
      strcat(fnInput, rootName);
      /* Open image file */
      if ((fpIn = freopenInputFile(fnInput, fpIn)) == NULL) {
	fprintf(stderr, "Can't open mean image file %s\n",fnInput);
	exit(0);
      }
      read_pfm_image(&meanImage, fpIn, fnInput, &nxImage, &nyImage); 
      nxBasis = nxImage;
      nyBasis = nyImage;
    }
  }
  
  fprintf(stderr, "Input basis image names (terminate with \"null\"):\n");
  nBasis =0;
  while(scanf("%s", fnBasis[nBasis]) != EOF) {
    if (checkForNull(fnBasis[nBasis]) == 0)
      break;
    nBasis++;
    if (nBasis >= MAXLEN) {
      fprintf(stderr, "Warning: Truncating basis to %d (MAXLEN) images.\n",
	      MAXLEN);
      break;
    }
  }
  if (nBasis == 0)
    error("No basis images.","");

  fprintf(stderr, "Input file name for singular values\n");
  sv = NULL;
  if (scanf("%s", rootName) != EOF) {
    strcpy(fnInput, pathNameInput);
    strcat(fnInput, rootName);
    /* Open image file */
    if ((fpIn = freopenInputFile(fnInput, fpIn)) == NULL) {
      fprintf(stderr, "Can't open singular value file %s\n",fnInput);
      exit(0);
    }
    /* Want 1 more singular value than basis functions to assess
       the contribution from the out of subspace error */
    grabFloatMemory(&sv, nBasis+1, "sv");
    for(k=0; k<=nBasis; k++) {
      if (fscanf(fpIn, "%e", sv+k) == EOF)
	error("Can't read singular values","");
    }
  }

  /* Input eigen basis images */
  fprintf(stderr, "Reading eigen basis images...");
  fpIn = NULL;
  for(kBasis=0; kBasis<nBasis; kBasis++) {
    strcpy(fnInput, pathNameInput);
    strcat(fnInput, fnBasis[kBasis]);
    /* Open image file */
    if ((fpIn = freopenInputFile(fnInput, fpIn)) == NULL) {
      fprintf(stderr, "Can't open image file %s\n",fnInput);
      exit(0);
    }
    read_pfm_image(&imageFlt, fpIn, fnInput, &nxImage, &nyImage); 
    if (kBasis == 0) {
      if (!use_mean) {
        nxBasis = nxImage;
        nyBasis = nyImage;
      } else {
	if ((nxImage != nxBasis) || (nyImage !=nyBasis)) {
	  fprintf(stderr, "Error: %s and %s different sizes\n",
		  fnBasis[0], fnBasis[kBasis]);
	  error("Bye","");
	}
      }
      size = nxBasis * nyBasis;
      for(c=0; c<nBasis; c++)
	grabFloatMemory( &(basisImages[c]), size, "basisImages");
    } else {
      if ((nxImage != nxBasis) || (nyImage !=nyBasis)) {
        fprintf(stderr, "Error: %s and %s different sizes\n",
		fnBasis[0], fnBasis[kBasis]);
        error("Bye","");
      }
    }
    /* Copy image to rows of basis array */
    for(i=0; i<nxBasis*nyBasis; i++) 
      basisImages[kBasis][i] = imageFlt[i];
    free(imageFlt);
  }
  fprintf(stderr, "done\n");
  grabFloatMemory(&coeff, nBasis, "coeff");
  xCent = ROUND(nxBasis/2.0);
  yCent = ROUND(nyBasis/2.0);
  /* Eyes aren't centered, set yCent to be the row for the eye center */
  yCent = ROUND(nyBasis * eyeUpY/(float)(eyeUpY+eyeDownY));

  /************************************************/
  /* Correlate eigen basis with each search image */
  /************************************************/
  for(kImage=0; kImage< nImage; kImage++) {

    /* Read pgm version of search image */
    strcpy(fnInput,fnImagePath[kImage]);
    strcat(fnInput,fnImage[kImage]);
    /* Open image file */
    if ((fpIn = freopenInputFile(fnInput, fpIn)) == NULL) {
      fprintf(stderr, "Can't open image file %s\n",fnInput);
      exit(0);
    }
    read_pgm_image(&image, fpIn, fnInput, &nxImage, &nyImage);
    size = nxImage*nyImage;
    if (nxImage<nxBasis || nyImage<nyBasis) {
      fprintf(stderr,"%s is smaller than eigen basis\n", fnInput);
    } else {

      grabFloatMemory(&residualImage, size, "residualImage");
      grabFloatMemory(&logLikeImage, size, "logLikeImage");
      for(ixy=0; ixy<size; ixy++) {
        residualImage[ixy] = 1.0;
        logLikeImage[ixy] = BIG_NEG;
      }

      /* Do basis projection correlation */
      if (!use_mean)
	error("Not using mean is not implemented yet.","");

      fprintf(stderr, "Correlating eigen basis on %s ...", fnImage[kImage]);

      corrEigenBasis(residualImage, logLikeImage, minLogLike,
		     meanImage,
		     basisImages, nxBasis, nyBasis, nBasis,
		     sv,
		     xCent, yCent,
		     image, nxImage, nyImage,
		     coeff);

      fprintf(stderr, " done\n");

      /* Built output image root name */
      pad_integer(numString, nBasis, 2);
      sprintf(rootName,"%s_n%s_%s", 
	      outputFileRoot, numString, fnImage[kImage]);
      rootName[strlen(rootName)-4] = '\0';  /* prune .pgm extension */

      range_float_image(residualImage, nxImage, nyImage, range);
      fprintf(stderr, " range of residuals: %f %f\n", range[0], range[1]);

      /* Save residual image */
      strcpy(outName, pathNameOutput);
      strcat(outName, rootName);
      strcat(outName, "_res");
      /*
      save_pfm_image(outName, residualImage, nxImage, nyImage);
      */

      range[0] = 0.0; range[1] = 1.0;
      out_of_range[0] = 0.0;  out_of_range[1] = 1.0;
      quantize_crop_range_float_image(&image, residualImage,
				      nxImage, nyImage, range, 
				      out_of_range, quant_par);
      save_pgm_image(outName, image, nxImage, nyImage);
      free(image);

      /* Output 1st principal component coeff image */

      crop[0] = BIG_NEG/2.0; crop[1] = -BIG_NEG;
      range_crop_float_image(logLikeImage, nxImage, nyImage, crop, range);
      fprintf(stderr, " range logLike image: %e %e\n", range[0], range[1]);
      crop[1] = max(range[1], minLogLike);
      crop[0] = max(range[1] - 1.0, minLogLike);  /* only 1 std dev down */
      out_of_range[0] = 0.0;  out_of_range[1] = 1.0;
      quantize_crop_range_float_image(&image, logLikeImage,
				      nxImage, nyImage,
				      crop, out_of_range, quant_par);
      strcpy(outName, pathNameOutput);
      strcat(outName, rootName);
      strcat(outName,"_logLike");
      save_pgm_image(outName, image, nxImage, nyImage);
      free(image);

      free(residualImage);
      free(logLikeImage);
    }
  }
  free(coeff);
  free(meanImage);
  for(c=0; c<nBasis; c++)
    free(basisImages[c]);
  return(0);
}

