/***
 FILE: mapping.h   June 1999
 Header file for mapping.c, utility routines for image warping. 

 Warp functions are used to define a new image as follows:
   newImage(x,y) = oldImage(warp(x,y; p))
 Here p = a vector of parameters.  The evaluation of oldImage
 between pixels is done by linear interpolation.

 WARNING:
  NO filtering is done during the image warping.  This should
  be included in order to control aliasing.  In particular, 
  the current warping code should not be used with warps involving
  significant (eg *2) scale changes in any direction.

 Notation:
  Images are represented by 1D (typically float) arrays, say image[ixy],
  where ixy = ix + iy * nx. Here (ix,iy) is the ix-th pixel along
  the iy-th scanline.
  Here (x,y) represents pixel coordinates, x to the left, y downwards,
  with the origin of an image taken at the top left corner.
  Image sizes are quoted as nx by ny, for nx pixels along each scan line
  and ny scan lines.
  In the comments, we use a simpler notation, namely image(x,y) or
  image(ix,iy) used to represent the image value at pixel location (x,y).
  If ix,iy are ints, then image(ix,iy) denotes image[ixy], with ixy as above.
  Otherwise, image(x,y) is denotes an interpolated value.

 The basic warps considered are:
   affine:
    affineWarp(x_vec; p) = x_origin + a_vec + (I + A) * dx
   with dx = x_vec - x_origin.   Here x_origin is the arbitrary
   origin for the coordinate transform, typically chosen to be
   the center of the image region during the warp estimation process. 

   rational:
    rationalWarp(x_vec; p) = x_origin + (a + A * dx)/ (1+b^T * dx)
   with dx = x_vec - x_origin.

   radial distortion (quadratic):
     radialWarp(x_vec; p) = x_0 + a_vec + dx * (1 + a0 * ||dx||^2)
   with dx = x_vec - x_0.  Here x_0 is the center of the radial
   distortion (typically the piercing point of the optical axis).
  
 Important operations on a warp function include:
  - translateWarp
      newImage(x,y) = oldImage(translateWarp(x,y; p, x0, y0))
                    = oldImage(warp(x+x0, y+y0; p)) 
    This allows the origin (0,0) for newImage to be redefined, effectively
    changing the crop box.  This is redundant for groups which include
    translation, but is included here for other warps such as the radial
    warp.

  - inverseWarp 
      (x,y) = inverseWarp( warp(x,y; p), p) 
    this defines the reverse mapping
      oldImage(x,y) = newImage(inverseWarp(x,y; p))

  - composeWarps
      composeWarps(x,y; p, q) = warp( warp(x,y; p); q)
    This is used, for example, for warps defined over image sequences
    where the warp from the first frame to the last involves a composition
    of the warps between successive frames.

 Both affine and rational warps form a group, with well defined
 composition and inverse mappings which can be represented within
 the group.  That is,
   inverseWarp(x,y; p) = warp(x,y; inverseWarpCoeffs(p))
   composeWarps(x,y; p, q) = warp(x,y; composeWarpCoeffs(p, q))
 The group property is very useful during the modelling of successive
 warps over several frames of an image sequence.

 It is also useful to consider the special case of translations,
 which are simply a redefinition of the origin in image space:
    translateWarp(x,y; p, x0, y0) = warp(x,y; translateWarpCoeffs(p, x0, y0))
 Translation can be implemented for any warp family, independently of
 whether it forms a group or not.

 These operations are implemented by the routines below.  Composition is
 only implemented for affine and rational warps.  

 TTD:
  Add filtering during warping in order to control aliasing.
l  
  Pass an image warp function to the warping routine(s), instead
  of using different routines for different types of warps.
***/

#ifndef _MAPPING_H

#define _MAPPING_H

#define MAX_MOTION_PAR 8  /* Maximum number of motion parameters */
#define ON_HORIZ  -1   /* Flag for point on horizon of rational warp */
#define OUTSIDE_HORIZ -2 /* Flag for point outside rational horizon */
#define INVALID_WARP_TYPE  -10  /* Flag for unrecognized warp type */

#define NUM_MAPPING_TYPES   4  /* Number of flavours of image mapping  */
/************  The warp number is determined by mappingType
  This is ordered by an inclusion relation.
const char *mappingType[NUM_MAPPING_TYPES] =
               {"translation", "rotation", "affine", "rational"};
*******************/
#define TRANSLATION 0
#define ROTATION    1
#define AFFINE      2
#define RATIONAL    3

typedef struct {
  char const * type; /* String label for warp type, one of:
	           translation, rotation, affine, rational */
  float x_origin[2];  /* spatial origin of warp */
  float par[MAX_MOTION_PAR];  /* current warp params */
  char frozen[MAX_MOTION_PAR];/* TRUE/FALSE flags indicating if parameter i 
				 is currently frozen */
  float sigma[MAX_MOTION_PAR];
  char frozenSigma[MAX_MOTION_PAR];
  int nPar; /* Number of warp parameters (frozen or not) */
  int nFree; /* Number of free warp parameters */
} ImageWarpStruct;

#endif

/******* Function prototypes ********/

/**  General interface routines **/

/* Given frame sample times t-baseDelay, and target times
      t-delay[0], t-delay[1], ...  t-delay[nTarget-1]
    for 0<=delay[k]<=maxDelay.
   Given the constant forward warp params, fwd, between each pair
   of frames t-k to t-(k-1), k<=maxDelay<=1
  OUTPUT:
    Compute the cummulative warps, 
       *cummWarp[k],  k=0,...,nTarget-1,
    from the base frame to the target frame delay[k], k=0,...,nTarget-1.

  EG: baseFrame = 0, nTarget = 3, maxDelay = 3;
         delay[k] = k+1,
    Then frames t-0,...,t-3 are considered. fwdWarp needs to be the
    warp from frame t-(k+1) to t-k, for k=0,1,2,3.
    On output cummWarp[k], k=0,..,2 gives the backwards
    warp from frame t-0 to target frame t-k, k = 1,..,3. 

  EG: baseFrame = 1, delay[*] = ( 0 2), nTarget = 2, maxDelay = 2 then 
    frame delay 1 is the base frame and cummWarp[0] maps this baseframe
    forward to the frame at time 0, and cummWarp[1] maps baseframe 1
    backwards to target frame 2. **/
void setCompositeWarps(ImageWarpStruct *cummWarp[], 
		      int baseDelay, int *delay, int nTarget,
		      ImageWarpStruct *fwdWarp, int maxDelay);

ImageWarpStruct * newImageWarpStruct();

void emptyImageWarpStruct(ImageWarpStruct *w);

int getImageWarpType(ImageWarpStruct * a);

char const * getImageWarpName(int type);


int convertImageWarpType(ImageWarpStruct *w, int toType);

char const * getImageWarpName(int type);

int promptImageWarp(ImageWarpStruct *warp);

/* Compute the warped image location of the pixel x[0-1], and
   return this location in xNew[0-1].
   Returns:
     INVALID_WARP_TYPE - if the warp is unrecognized.
     0 if successful
     Other (non-zero) warp specific flags, such as ON_HORIZ for
     rational warps.
*/
int ptSampleImageWarp(float *x, ImageWarpStruct *warp, float *xNew);

/* Compute the jacobian of the warped image location with respect
   to the free warp parameters, sampled at image position x,
   and return this 2 x (warp->nPar) matrix in jac.
   Returns:
     INVALID_WARP_TYPE - if the warp is unrecognized.
     0 if successful
     Other (non-zero) warp specific flags, such as ON_HORIZ for rational warps.
*/
int derivSampleImageWarp(float *x, ImageWarpStruct *warp,
			 float jac[2][MAX_MOTION_PAR]);

/* Compose warps represented by warp structs.
   The composed warp is returned in *pResult,
        result(x) = second( first (x))
   Returns 0 if successful.
*/
int composeImageWarps(ImageWarpStruct *first, ImageWarpStruct *second, 
		      ImageWarpStruct **pResult);

int moveImageWarpOrigin(ImageWarpStruct *a, float *newOrig, 
			ImageWarpStruct **pResult);

int invertImageWarp(ImageWarpStruct *a, ImageWarpStruct **pResult);

int convertImageWarpType(ImageWarpStruct *w, int toType);

/* Delete the columns of the image warp jacobian for the frozen parameters */
void compressWarpJac(float jac[2][MAX_MOTION_PAR], int nPar, char *frozen); 

/* Expand a compressed vector of the unfrozen (free) warp parameters
   to the standard form.  Frozen parameters are left as zero in
   this expanded form, which allows for a simple update.
   The result is given in fullPar.  The arrays *freePar and *fullPar
   can be identical storage locations. */
void expandWarpParams(float *freePar, int nPar, int nFree, char *frozen, 
		      float *fullPar);

/***
 * Do a linear interpolation between low_value and
 * high_value, for position going from 0 to 1, respectively.
 * Returns:
 *  result = low_value * (1 - position) + high_value * position
 ***/
float linear_interp(float low_value, float high_value, float position);

/* ================ Affine 2D Transforms ==================== **
  Definition:
    Affine Transform:
      x_new = affine(x_old; x_origin, a) 
            = x_origin + a_vec + (I + A) (x_old - x_origin)  (Eqn 1a)
            = x_old + a_vec + A (x_old - x_origin)           (Eqn 1b)
    with: 
      x_old - a 2 vector for the pixel coords of the point to be transformed
      x_new - the warped pixel coordinates
      x_origin = the origin for the definition of the affine transform,
               = (x_cent, y_cent)
      a_vec and A the six affine parameters.

    To represent this transform, a float array a[6] is used, with
    the coefficients representing the following values:
      a[6] = (a_vec[0], a_vec[1], A11, A21, A12, A22)

    This is a redundant representation, if you change x_origin then
    an equivalent affine transformation exists with different values
    for a_vec and A (see movAffineOrigin()).

    Note that the image displacement, namely
     d = x_new - x_old,
    is given by:
     d = a_vec + A * (x_old - x_origin)

*================ End: Defn Affine 2D Transform ================*/

/*****
  Affine transform procedures:

  affinePtSample:
    Sample the affine transform 
      x_new = affine(x_old; x_origin, aff)
    with
       float x_origin[2], float aff[6], 
    defined as above.

    Input: 
       x = x_old[0]
       y = x_old[1]

       aff = the 6-d affine parameter vector (a_vec, A) as above.
    On return:
      x_new = affine(x_old; x_origin, aff);
    Returns value: 0 (always) for consistency with other warp sampling
                      routines such as rationalPtSample().
****/
int affinePtSample(float *x_old, float *x_orig, float *aff, float *x_new);

/* Return the jacobian of the affine transform with respect to
   the paramters, sampled at image position x_old */
int affineJac(float *x_old, float *x_orig, float *aff, 
	      float jac[2][MAX_MOTION_PAR]);

/****
  affineDispSample: 
    Return the image displacement, float d[2], for pixel x_old, 
    according to the affine transform defined by 
       float x_origin[2], float a[6], as above.
    Input: 
       x_diff = x_old - x_origin (a 2-vector)
       aff = the 6-d affine parameter vector (a_vec, A).
    On return:
      displacement d = affine(x_old; x_origin, a) - x_old;
                     = a_vec + A * x_diff   (see Eqn 1b above)
    Return value: 0 (always successful).
****/
int affineDispSample(float *x_diff, float *aff, float *d);

/**** 
  composeAffine:
    Evaluate composition of affine transforms a and b, the
    composite transform is returned in a,
      a <- b o a 
    The origin of the returned composite transform is x_origA
****/
void composeAffine(float *x_origA, float *a, float *x_origB, float *b);

/****
  moveAffineOrigin:
   Move the origin of the affine map from x_origOld to x_origNew
   Coefficients a are updated.
****/
void moveAffineOrigin(float *x_origOld, float *a, float *x_origNew);

/****
  invertAffine:
    Compute Inverse Affine Image Transform :
    Affine Transform (Eqn 1a)
       x_new = x_origin + a_vec + (I + A) (x_old - x_origin)
    Inverse Affine Transform:
     x_old = x_orig + b + (I + B) (x_new - x_orig)
    with
     B = inverse(I+A) - I
     b = -(I+B) * a

   Subroutine:
     int invertAffine(a, b)
   
     Input:  Float affine parameters
      a = ( a_0, a_1, A_00, A_10, A_01, A_11)
      (NOTE: Order is down columns first)

     Returns:
      0 if successful
     -1 if failure due to singularity of input transformation.

     Output:  Float affine parameters
      b = ( b_0, b_1, B_00, B_10, B_01, B_11)
     (NOTE: Order is down columns first)
****/
int invertAffine(float *a, float *b);

/* ============ Rational 2D Transforms ============== */

/****
  rationalPtSample:
    Compute sample value, x_new, of rational image transform 
      x_new = x_origin + (a + A(x_vec-x_origin)/(1+b^T (x_vec-x_origin))
    where x_origin = origin for transform.  Here
    x_vec = (x, y)^T, and similarly for x_origin.
   
    Returns:
     ON_HORIZ if the denominator within MACHINE_EPS of 0.0, returns -1
     OUTSIDE_HORIZ if the denominator < -MACHINE_EPS
      0 otherwise (success).
****/
int rationalPtSample(float *x_old, float *x_orig, float *par, float *x_new);

/* Return the jacobian of the rational transform, with respect to
   the paramters, at image position x_old */
int rationalJac(float *x_old, float *x_orig, float *par, 
	      float jac[2][MAX_MOTION_PAR]);

/***
  invertRational:
    Given rational transform:
      x'-x0 = (a + A(x-x0)/(1+b^T (x-x0))
    where x0 = origin for transform,
    compute inverse
      x-x0 = (a' + A'(x-x0))/(1+b'^T (x-x0))
    
    Here the original transform is given in terms of

    a[0-7] = (a_1 a_2 A_11 A_21 A_12 A_22 b_1 b_2)
    indices.. 0   1   2    3    4    5    6   7

    and similarly, the inverse transform is returned
    in  b[0-7]

    Returns value 0 if successful, else -1 if transform is singular
***/
int invertRational(float *a, float *b);

/**** 
  composeRational:
    Evaluate composition of affine transforms r and d, the
    composite transform is returned in r,
      r <- d o r 
    The origin of the returned composite transform is x_origR
****/
void composeRational(float *x_origR, float *rFlt,
		     float *x_origD, float *dFlt);

/* move the origin of the rational map a from x_orig0 to x_orig1 */
void moveRationalOrigin(float *x_orig0, float *a, float *x_orig1);

/* Map orientation vector, inVec, at x to new orientation, outVec,
   according to rational mapping having coefficients x_origin, a */
void mapRationalVec(float *inVec, float *x, 
		    float *x_orig, float *a, 
		    float *outVec);

/* Given a special rational warp defined in *warp, check to see if it is of a
   recognized sub type of the rational warp (return 0, else return -1).
   If it is recoginzed and:
     If toLevel>=1:
        unpack the warp as an equivalent rational warp.
     If toLevel>=2:
        set the jacobian, partial rat/ partial par.
*/   
int setRationalWarp(ImageWarpStruct *warp, int toLevel); 

/* Change set level of warp to uninitialized level */
int touchRationalWarp(ImageWarpStruct *warp);

/* ============ Radial distortion ============== */
/**** 
  radialPtSample:
    Sample the radial distortion map
      x_new = (x_vec - x_origin) * (1 + d0 * ||x_vec - x_origin||^2)
               + x_origin + shift
    That is, given x_vec and the parameters x_origin, shift, d0,
    then find x_new.
    Here array a[3] = (shift_x, shift_y, d0).
    Returns 0 (always successful).
****/
int radialPtSample(float *x, float *x_orig, float *a, float *x_new);

/**** 
  inverseRadialPtSample:
    Sample the inverse of the radial distortion map
      x_vec = (w - x_origin) * (1 + d0 * ||w - x_origin||^2) + x_origin + shift
    That is, given x_vec and the parameters x_origin, shift, a0,
    then find w.
    Here array a[3] = (shift_x, shift_y, d0).

    Returns 0 if successful, else -1 (failed to solve for inverse point).
****/
int inverseRadialPtSample(float *x, float *x_orig, float *a, float *w);

/* ============ Image Warping ============== */

/*
 Backwards warp of a nx by ny image by an affine warp with parameters
 given by (x_origin, b).  
 That is, for pixel (x,y):
   warpedImage(x,y) =(apprx) origImage(affine(x,y; x_origin, b))
 where affine(x,y; x_origin, b) gives the pixel coords for
 the (backwards) affine warp evaluated at (x,y).

 The result is stored in the float array warped, which must be
 already allocated to hold at least an nx * ny array.

 Pixel values are determined by linear interpolation from the original image.

 While this is fast and simple, a better interpolation
 may be desirable.

 WARNING: Filtering would be essential if the image is to be
 subsampled significantly.
*/
void affine_warp_image(float *image, int nx, int ny,
		       float *warped, 
		       float *x_orig, float *b, 
		       unsigned char defaultGrey);

/*
Backwards warp of "image" by warp b.  The new image is given by
 temp(x,y) = image(affine(x,y; x_origin, b))
This image is sampled at pixels:
 (x,y) = (xMin, yMin) + (dx * i, dy * j) 
to form the float image warped(i,j).

This is similar to affine_warp_image() above, but
here the warped image can be resampled at a different rate.

Pixel values are determined by linearly interpolating the warped image.
No filtering is done.  This is needed for some warps in order to control
aliasing. */
void affine_warp_sample_image(float *image, int nx, int ny,
                  float *warped, 
                  float xMin, float yMin, float dx, float dy, 
                  int newNx, int newNy,
                  float *x_orig, float *b, 
		  unsigned char defaultGrey);

/*
Backwards warp of "image" by rational warp b.
The result is stored in "warped".
Pixel values are determined by linearly interpolating the warped image.
While this is fairly fast, a better interpolation may be desirable.
*/
void rational_warp_image(float *image, int nx, int ny,
			 float *warped, 
			 float *x_orig, float *b,
			 unsigned char defaultGrey);

/*
Backwards warp of "image" according to radial warp.
The result is stored in "warped".
Pixel values are determined by linearly interpolating the warped image.
While this is fairly fast, a better interpolation may be desirable.
*/
void radial_warp_image(float *image, int nx, int ny,
		       float *warped,
		       float *x_orig, float *par, 
		       unsigned char defaultGrey);

/****
 warpAndBound:
  Forwards warp of "image" by rational warp a.
  The result is stored in "warped".
  The size of warped is the smallest upright rectangle containing
  the mapping of "image".  The bounding box is restricted to lie on
  a subsampled grid, starting at 0,0 with spacing `subSample'.
  The size of warped is returned as *pnxw, *pnyw.
  Grey levels are determined by linearly interpolating the warped image.
  While this is fairly fast, a better interpolation may be desirable.

  Returns the backwards warp from new warped image coords to original
  image in b with center x_origin left unchanged.  That is:
     warped(x,y) = image(rationalWarp(x,y; x_orig, b))
****/
void warpAndBound(float *image, int nx, int ny, 
		  float **pwarped, int *pnxw, int *pnyw, 
		  float *x_orig, float *a, 
		  unsigned char defaultGrey,
		  int subSample, float *b);

/****
 warpCrop:
  Forwards warp of "image" by rational warp a.
  The result is stored in "warped".

  The size of warped cropped to bounding box with top left
  corner ix0,iy0, and bottom right corner ixy, iy1.
  Grey levels are determined by linearly interpolating the warped image.
  While this is fairly fast, a better interpolation may be desirable.
****/
void warpCrop(float *image, int nx, int ny, 
	      float **pwarped, 
              int ix0, int iy0, int ix1, int iy1, 
	      float *x_orig, float *a,
              unsigned char defaultGrey);


