# include "string.h"
# include "stdio.h"
# include "stdlib.h"

# include "utils.h"

void error(const char *s1, const char *s2)
 {
  fprintf(stderr, "Error: %s %s\n", s1, s2);
  exit(0);
 }

int checkForNull(fname)  /** Returns 0 when fname is Null, NULL, null, etc. **/
 char *fname;
 {
  if (strcasecmp(fname, "NULL") == 0)
   return((int) 0);
  else
   return((int) 1);
 }
 
void grabFloatMemory(x,n,blurb)
 float **x;
 int n;
 char *blurb;
 {
   if ((*x = (float *) malloc((size_t) n * sizeof(float)) ) == NULL)
     error(" Unable to allocate memory for ", blurb);
 }
 
void grabByteMemory(x,n,blurb)
 char **x;
 int n;
 char *blurb;
 {
   if ((*x = (char *) malloc((size_t) n * sizeof(char)) ) == NULL)
     error(" Unable to allocate memory for ", blurb);
 }
 
void grabIntMemory(x,n,blurb)
 int **x;
 int n;
 char *blurb;
 {
   if ((*x = (int *) malloc((size_t) n * sizeof(int)) ) == NULL)
     error(" Unable to allocate memory for ", blurb);
 }

/* input a single line from a file (up to the character '\n' or
   maxSize-1 characters, whichever comes first).  
   getLine returns:
     1    if maxSize-1 is reached (line continues beyond buffer)
     0    if the carriage return is read (end of line read)
    -1    if eof is bumped into or could not read one character from file.
************************************************************************/
int getLine(FILE *fpIn, char *line, int maxSize)
 {
  int k, endLine;
  char *p;
  endLine = 0;
  for(p=line, k=0; k<maxSize-1; k++, p++) {
   if (!fread(p, (int) 1, 1, fpIn)) {
    *p = '\0';
    return(-1);
   }
   if (*p == '\n') {
    endLine = 1;
    *p = '\0';
    break;
   }
  }
  if (endLine == 0) {
   line[maxSize-1] = '\0';
   return(1);
  } else
   return(0);
 }

/* input a single line from a file (up to the character '\n' or
   maxSize-1 characters, whichever comes first).  
   Strip leading blanks and tabs.
   getStripLine returns:
     1    if maxSize-1 is reached (line continues beyond buffer)
     0    if the carriage return is read (end of line read)
    -1    if eof is bumped into or could not read one character from file.
************************************************************************/
int getStripLine(FILE *fpIn, char *line, int maxSize)
 {
  int k, endLine, foundNonWht;
  char *p;
  endLine = 0;
  foundNonWht = 0;
  
  for(p=line, k=0; k<maxSize-1; k+=foundNonWht, p+=foundNonWht) {
   if (!fread(p, (int) 1, 1, fpIn)) {
    *p = '\0';
    return(-1);
   }
   if (*p == '\n') {
    endLine = 1;
    *p = '\0';
    break;
   }
   if (!foundNonWht) {
    /* Check for first appearance of non blank or tab */
    if (!(*p == ' ' || *p == '\t'))
     foundNonWht = 1;
   }
  }
  if (endLine == 0) {
   line[maxSize-1] = '\0';
   return(1);
  } else
   return(0);
 }

/* Flush a line from an input file (up to the character '\n' or EOF)
   flushLine returns:
    0 if successful (character '\n' found)
   -1 if eof bumped into or cannot read character.
*************************************************************/
int flushLine(FILE *fpIn)
 {
  char p;
  do {
   if (!fread(&p, (int) 1, 1, fpIn))
    return(-1);
  } while (p != '\n');
  return(0);
 }

/* Read next integer in input file, ignoring comments
   which are indicated by # and continue to the next newline.
   Returns 1 (success), 0 (failure),  else EOF */
int readOneInt(FILE *fp, int *pn)
{
 char comment[MAXLEN];
 int more, numRead;
 
 do {
  numRead = fscanf(fp, "%d", pn);
  if (numRead != 1) { /* Strip expected comment */
   more = getStripLine(fp, comment, MAXLEN);
   if (more == EOF) {
     return(EOF);
   }
   if (comment[0] == '#') /* Comment beginning detected */
     if (more == 1) {
       if (flushLine(fp) == EOF)
         return(EOF);
     }
   } else { /* Garbled input */
     return(0);
   }
 } while( numRead != 1);

 return(numRead);
}

/* Read next float in input file, ignoring comments
   which are indicated by # and continue to the next newline 
   Returns 1 (success), 0 (failure),  else EOF */
int readOneFloat(FILE *fp, float *px)
{
 char comment[MAXLEN];
 int more, numRead;
 
 do {
  numRead = fscanf(fp, "%e", px);
  if (numRead != 1) { /* Strip expected comment */
   more = getStripLine(fp, comment, MAXLEN);
   if (more == EOF) {
     return(EOF);
   }
   if (comment[0] == '#') /* Comment beginning detected */
    if (more == 1)
      if (flushLine(fp) == EOF)
       return(EOF);
   } else { /* Garbled input */
     return(0);
   }
 } while( numRead != 1);

 return(numRead);
}


/* Read next string in input file, ignoring comments
   which are indicated by # and continue to the next newline.
   It is assumed the string is long enough to contain the input. 
   Returns 1 (success), else EOF */
int readOneString(FILE *fp, char *s)
{
 char comment[MAXLEN];
 int more, numRead;
 
 do {
  numRead = fscanf(fp, "%s", s);
  if (numRead != 1) 
    return(EOF);
  if (s[0] == '#') { /* Strip comment */
   numRead = 0;
   if (flushLine(fp) == EOF)
     return(EOF);
  }
 } while( numRead != 1);

 return(numRead);
}

/**********************************************
Sort float array w into decreasing order, using insertion sort
Return permutation array iPerm *... ith sorted value is w[iPerm[i]] 
***********************************************/
void sortFloatArray( int *iPerm, float *w, int numSamp)
{
  float tmp;
  int i, j, iTmp;
   
  for(i=0; i<numSamp; i++)
    iPerm[i] = i;

  for(i=0; i<numSamp-1; i++) {
    /* find max in remaining set i..numSamp */
    tmp = w[iPerm[i]];
    iTmp = i;
    for(j=i+1; j<numSamp; j++)
      if (w[iPerm[j]] > tmp) {
	tmp = w[iPerm[j]];
	iTmp = j;
      }
    /* Switch */
    j = iPerm[i];
    iPerm[i] = iPerm[iTmp];
    iPerm[iTmp] = j;
  }
}

